<?php

namespace FiloBlu\Flow\Model\Channel\In;

use Magento\Framework\Exception\FileSystemException;
use Magento\ImportExport\Model\Import\Adapter as ImportAdapter;
use Magento\Framework\App\Filesystem\DirectoryList;
use Magento\ImportExport\Model\Import;
use Magento\Framework\App\ObjectManager;
use Magento\Catalog\Model\Product;
use FiloBlu\Flow\Model\Imagesflow;
use Magento\MediaStorage\Model\File\Uploader;

/**
 * Class Images
 * @package FiloBlu\Flow\Model\Channel\In
 */
class Images extends AbstractModel
{
    /**
     * @var string
     */
    const STATUS_RECEIVED = 'received';

    /**
     * @var string
     */
    const STATUS_PROCESSED = 'processed';

    /**
     * @var string
     */
    const STATUS_PARSED = 'parsed';

    /**
     * @var string
     */
    const STATUS_ERROR = 'error';

    /**
     * @var \FiloBlu\Flow\Helper\Monolog\Logger
     */
    protected $_logger;

    /**
     * @var \FiloBlu\Flow\Helper\Images
     */
    protected $_imagesHelper;
    /**
     * @var \Magento\Framework\App\ObjectManager $_objectManager
     */
    protected $_objectManager;
    /**
     * @var \Magento\Framework\App\ResourceConnection
     */
    protected $_resourceConnection;
    /**
     * @var \Magento\Framework\DB\Adapter\AdapterInterface
     */
    protected $_connection;
    /**
     * @var \Magento\Catalog\Model\ProductRepository
     */
    protected $_productRepository;
    /**
     * @var \Magento\Framework\Filesystem
     */
    protected $_filesystem;
    /**
     * @var \Magento\Framework\Filesystem\Directory\WriteInterface
     */
    protected $_mediaDirectory;
    /**
     * @var \Magento\Framework\Filesystem\Directory\WriteInterface
     */
    protected $_varDirectory;
    /**
     * @var  \Magento\Catalog\Model\Product\Image\CacheFactory
     */
    protected $_imageCacheFactory;
    /**
     * @var array
     */
    protected $_lookupAttributeSet = [];

    /**
     * @var
     */
    private $channel_photo_config;
    /**
     * @var
     */
    private $collection_of_images_to_parse;
    /**
     * @var
     */
    private $array_of_data_for_csv;
    /**
     * @var string
     */
    private $filename_separator = '_';
    /**
     * @var string
     */
    private $image_alt_separator = ' ';
    /**
     * @var \Magento\Framework\Event\Manager
     */
    private $eventManager;

    /**
     * @var int
     */
    private $_errorsProcessingImages = 0;

    /**
     * @var int
     */
    private $_imagesToProcess = 0;

    /**
     * @var string
     */
    private $unique_image_code_pattern = 'unq';

    /**
     * @var string
     */
    private $cache_folder = BP . '/pub/media/catalog/product/cache/';

    /**
     * Product entity link field
     *
     * @var string
     */
    private $productEntityLinkField;

    /**
     *
     * TODO: check locks on import
     * TODO: vendor/magento/module-catalog-import-export/Model/Import/Product.php see how to use this class
     */
    /**
     * Images constructor.
     * @param \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider
     * @param \FiloBlu\Flow\Model\Channel\ConfigFactory $channelConfigFactory
     * @param \FiloBlu\Flow\Helper\Images $imagesHelper
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Framework\Event\Manager $eventManager
     * @param \Magento\Framework\App\ResourceConnection $resourceConnection
     * @param \Magento\Catalog\Model\ProductRepository $productRepository
     * @param \Magento\Framework\Filesystem $filesystem
     * @param Product\Image\CacheFactory $imageCacheFactory
     */
    public function __construct(
        \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider,
        \FiloBlu\Flow\Model\Channel\ConfigFactory $channelConfigFactory,
        \FiloBlu\Flow\Helper\Images $imagesHelper,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Framework\Event\Manager $eventManager,
        \Magento\Framework\App\ResourceConnection $resourceConnection,
        \Magento\Catalog\Model\ProductRepository $productRepository,
        \Magento\Framework\Filesystem $filesystem,
        \Magento\Catalog\Model\Product\Image\CacheFactory $imageCacheFactory
    )
    {
        parent::__construct($loggerProvider, $channelConfigFactory);

        $this->_logger = $loggerProvider->getLogger();
        $this->_imagesHelper = $imagesHelper;
        $this->eventManager = $eventManager;

        $this->_objectManager = ObjectManager::getInstance();
        $this->_resourceConnection = $resourceConnection;
        $this->_connection = $resourceConnection->getConnection();

        $this->_productRepository = $productRepository;
        $this->_filesystem = $filesystem;
        $this->_mediaDirectory = $filesystem->getDirectoryWrite(DirectoryList::MEDIA);
        $this->_varDirectory = $filesystem->getDirectoryWrite(DirectoryList::VAR_DIR);
        $this->_imageCacheFactory = $imageCacheFactory;

        $this->loadLookupTables();
    }

    /**
     * @return int
     */
    public function getProcessedImageErrorCount()
    {
        return $this->_errorsProcessingImages;
    }

    /**
     * @return int
     */
    public function getImageToProcessCount()
    {
        return $this->_imagesToProcess;
    }

    /**
     * Load the lookups tables to speed up the whole process
     */
    private function loadLookupTables()
    {
        $stmt = $this->_connection->query('SELECT a.attribute_set_id, b.* FROM eav_attribute_set AS a LEFT JOIN eav_entity_type AS b ON b.entity_type_id = a.entity_type_id WHERE attribute_set_id IN ( SELECT DISTINCT attribute_set_id FROM catalog_product_entity )');

        $results = $stmt->fetchAll(\Zend_Db::FETCH_OBJ);

        foreach ($results as $item) {
            $item->attributesSet = $this->getAttributeSet($item->entity_type_id);
            $this->_lookupAttributeSet[$item->attribute_set_id] = $item;
        }
    }

    /**
     * Retrieve the attribute set available from the entity type
     *
     * @param $entity_type_id
     * @return array
     */
    private function getAttributeSet($entity_type_id)
    {
        $attributeSet = [];

        $stmt = $this->_connection->query(
            'SELECT * FROM eav_attribute WHERE entity_type_id = ?',
            [$entity_type_id]
        );

        $results = $stmt->fetchAll(\Zend_Db::FETCH_OBJ);

        foreach ($results as $item) {
            $attributeSet[$item->attribute_code] = $item;
        }

        return $attributeSet;
    }

    /**
     * @param $channelname
     */
    public function load($channelname)
    {
        $this->init($channelname);
    }

    /**
     * @param $file
     * @return bool
     */
    public function insertData($file)
    {
        $meta_file_id = $file->getId();

        /** @var \FiloBlu\Flow\Model\Imagesflow $model */
        $model = $this->_objectManager->create(\FiloBlu\Flow\Model\Imagesflow::class);

        // Loading images SKUs to process
        $collection = $model->getCollection();
        $collection->addFieldToSelect(['id', 'image_sku']);
        //$collection->addFieldToFilter('meta_processed', 0);
        $collection->addFieldToFilter('meta_file', $meta_file_id);
        $collection->addFieldToFilter('image_status', ['in' => [Imagesflow::STATUS_PARSED, Imagesflow::STATUS_ERROR]]);
        $collection->getSelect()->group('image_sku');
        $collection->load();

        $this->_imagesToProcess = $collection->count();

        /** @var \Magento\Framework\Model\ResourceModel\Iterator $iterator */
        $iterator = $this->_objectManager->create(\Magento\Framework\Model\ResourceModel\Iterator::class);
        $iterator->walk($collection->getSelect(), [[$this, 'fastImageProcess']], ['meta_file_id' => $meta_file_id]);

        if ($this->_errorsProcessingImages > 0)
        {
            return false;
        }
        else
        {
            return true;
        }
    }

    /**
     * Process image import directly query the db layer
     *
     * @param $args
     * @return bool
     */
    public function fastImageProcess($args)
    {
        /* Where Magento 2 looks for images */
        $catalogRootDirectory = $this->_mediaDirectory->getAbsolutePath();

        /* Retieve sku */
        $sku = $args['row']['image_sku'];
        $meta_file_id = $args['meta_file_id'];

        /* Find product by sku */

        $results = $this->_connection->query('SELECT * FROM catalog_product_entity WHERE sku = ?', [$sku]);

        $product = $results->fetch(\Zend_Db::FETCH_OBJ);

        if ($product === null || $product === false)
        {
            $this->_logger->error("Unable to find product with SKU {$sku}");
            $this->_errorsProcessingImages++;
            return true;
        }

        /* TODO : use direct query instead. See https://blackfire.io/profiles/0b7ca6a2-ac40-41c1-b41d-1bfee4e7a8ac/graph*/
        
        /* Find incoming product images */

        /**
         * @var Imagesflow resource
         */
        $model = $this->_objectManager->create(\FiloBlu\Flow\Model\Imagesflow::class);

        $collection = $model->getCollection();
       // $collection->addFieldToFilter('meta_processed', 0);
        $collection->addFieldToFilter('meta_file', $meta_file_id);
        $collection->addFieldToFilter('image_status', ['in' => [Imagesflow::STATUS_PARSED, Imagesflow::STATUS_ERROR]]);
        $collection->addFieldToFilter('image_sku', $sku);
        //$collection->setOrder('image_position', 'ASC');
        $collection->load();

        foreach ($collection as $imageFlow)
        {
            $incomingImages[Uploader::getCorrectFileName($imageFlow->getImageName())] = $imageFlow;
        }

        $id = intval($product->attribute_set_id);
        $attributeSet = $this->_lookupAttributeSet[$id];

        $mediaGaleryAttribute = $attributeSet->attributesSet['media_gallery'];
        $mediaGaleryAttributeId = $mediaGaleryAttribute->attribute_id;

        /* Fix $catalogRootDirectory */

        $catalogRootDirectory .= str_replace('_', DIRECTORY_SEPARATOR, $attributeSet->entity_type_code);

        /* Find existing images in gallery */
        $existingImages = $this->getProductGalleryBySku($sku, $mediaGaleryAttributeId);

        foreach ($incomingImages as $name => $imageFlow)
        {
            try
            {
                $relativePath = Uploader::getDispretionPath($name) . DIRECTORY_SEPARATOR . $name;
                $to = $catalogRootDirectory . $relativePath;
                $from = $this->getImportImagePath($imageFlow);

                $folder = dirname($to) . DIRECTORY_SEPARATOR;

                if (false === is_dir($folder))
                {
                    if(false === mkdir($folder, 0777, true))
                    {
                        $this->_logger->error("Unable to create directory {$folder}");
                        $this->setProcecessedStatus($imageFlow, Imagesflow::STATUS_ERROR, "Unable to create directory {$folder}");
                        $this->_errorsProcessingImages++;
                        return true;
                    }
                }

                if(false === copy($from, $to))
                {
                    $this->setProcecessedStatus($imageFlow, Imagesflow::STATUS_ERROR, "Unable to move {$from} -> {$to}");
                    $this->_errorsProcessingImages++;
                }
                /* If already esists image in gallery just copy/move the image file and update metas */
                if (array_key_exists($name, $existingImages)) {
                    $row = $existingImages[$name];
                    $this->updateGalleryValues($product->row_id, $row->value_id, $imageFlow);
                    $this->updateAttributes($product->row_id, $relativePath, $imageFlow, $attributeSet->attributesSet);
                } else {
                    $this->addImageToGallery($relativePath, $product->row_id, $mediaGaleryAttributeId, $imageFlow);
                    $this->addAttributes($product->row_id, $relativePath, $imageFlow, $attributeSet->attributesSet);
                }

                $this->setProcecessedStatus($imageFlow, Imagesflow::STATUS_PROCESSED);
            } catch (\Exception $e) {
                $this->_logger->info("Importing image {$name} for product with SKU {$sku} failed. : " . $e->getMessage() . PHP_EOL . $e->getTraceAsString());
                $this->setProcecessedStatus($imageFlow, Imagesflow::STATUS_ERROR, $e->getMessage());
                $this->_errorsProcessingImages++;
            }
        }

        return true;
    }

    /**
     * Retrieve the whole gallery for the given product sku
     * @param $sku
     * @param $media_gallery_attribute_id
     */
    private function getProductGalleryBySku($sku, $media_gallery_attribute_id)
    {
        $gallery = [];

        $results = $this->_connection->query(
            'SELECT CPEMGON.*, CPEMGVTE.row_id FROM catalog_product_entity AS CPE
             LEFT JOIN catalog_product_entity_media_gallery_value_to_entity AS CPEMGVTE ON CPE.row_id = CPEMGVTE.row_id
             LEFT JOIN catalog_product_entity_media_gallery AS CPEMGON ON CPEMGVTE.value_id = CPEMGON.value_id
             WHERE CPE.sku = ? AND CPEMGON.attribute_id = ?', [$sku, $media_gallery_attribute_id]);

        $images = $results->fetchAll(\Zend_Db::FETCH_OBJ);

        foreach ($images as $image) {
            /* Remove the dipersion prefix */
            $gallery[basename($image->value)] = $image;
        }

        return $gallery;
    }

    /**
     * @param Imagesflow $imagesflow
     * @return string
     * @throws FileSystemException
     */
    protected function getImportImagePath(Imagesflow $imagesflow)
    {
        return $this->_varDirectory->getAbsolutePath() . DIRECTORY_SEPARATOR . $imagesflow->getImagePath() . DIRECTORY_SEPARATOR . $imagesflow->getImageName();
    }

    /**
     * @param $row_id
     * @param $value_id
     * @param Imagesflow $imageFlow
     */
    private function updateGalleryValues($row_id, $value_id, Imagesflow $imageFlow)
    {
        $valueTable = $this->_connection->getTableName('catalog_product_entity_media_gallery_value');

        $where =  $this->_connection->quoteInto('value_id = ? AND row_id = ?', $value_id, $row_id);

        $this->_connection->update($valueTable,
            [
                'store_id' => 0,
                'label' => $imageFlow->getImageAlt(),
                'position' => $imageFlow->getImagePosition(),
                'disabled' => $imageFlow->getImageHide()
            ],
           $where);
    }

    /**
     * Update attributes. Allowed attributes to be set are defined in @see \FiloBlu\Flow\Model\Channel\In\Images::$_imageAttributeCodes
     *
     * @param $row_id
     * @param $value
     * @param Imagesflow $imageFlow
     * @param array $attributesSet
     */
    protected function updateAttributes($row_id, $value, Imagesflow $imageFlow, array $attributesSet)
    {
        $assignableAttributes = explode('|', $imageFlow->getImageType());

        if (empty($assignableAttributes)) {
            return;
        }

        $attributesTable = $this->_connection->getTableName('catalog_product_entity_varchar');

        $attributeIds = [];
        $toUpdate = [];
        $lookup = [];

        foreach ($assignableAttributes as $attributeCode) {
            if (!isset($attributesSet[$attributeCode])) {
                continue;
            }

            $id = ($attributesSet[$attributeCode])->attribute_id;
            $attributeIds[] = intval($id);
            $lookup[$id] = $attributeCode;
        }

        $stmt = $this->_connection->query(
            $this->_connection
                ->select()
                ->from($attributesTable)
                ->where('row_id = ?', $row_id)
                ->where('store_id = ?', 0)
                ->where('attribute_id IN (?)', $attributeIds)
        );

        $existingAttributes = $stmt->fetchAll(\Zend_Db::FETCH_OBJ);

        /* Find attributes to update and attributes to insert */

        foreach ($existingAttributes as $attribute) {
            $id = $attribute->attribute_id;
            $toUpdate[] = $lookup[$id];

            $where =  $this->_connection->quoteInto('value_id = ?', $attribute->value_id);

            $this->_connection->update($attributesTable,
                [
                    'store_id' => 0,
                    'attribute_id' => $id,
                    'row_id' => $row_id,
                    'value' => $value
                ], $where);
        }

        /* Add missing attributes */

        foreach (array_diff($assignableAttributes, $toUpdate) as $attributeCode) {

            if (!isset($attributesSet[$attributeCode])) {
                continue;
            }

            $attributeId = ($attributesSet[$attributeCode])->attribute_id;

            $affectedRows = $this->_connection->insert($attributesTable, [
                    'store_id' => 0,
                    'attribute_id' => $attributeId,
                    'row_id' => $row_id,
                    'value' => $value
                ]);
        }
    }

    /**
     * @param $value
     * @param $row_id
     * @param $attribute_id
     * @param Imagesflow $imageFlow
     */
    private function addImageToGallery($value, $row_id, $attribute_id, Imagesflow $imageFlow)
    {
        $galleryTable = $this->_connection->getTableName('catalog_product_entity_media_gallery');
        $relationTable = $this->_connection->getTableName('catalog_product_entity_media_gallery_value_to_entity');
        $valueTable = $this->_connection->getTableName('catalog_product_entity_media_gallery_value');

        /* Add new entry in the gallery */

        $affectedColumns = $this->_connection->insert($galleryTable, [
            'value' => $value,
            'attribute_id' => $attribute_id,
            'media_type' => 'image',
            'disabled' => 0
        ]);

        /* Create relationship */
        $value_id = $this->_connection->lastInsertId($galleryTable);

        $this->_connection->insert($relationTable, [
            'value_id' => $value_id,
            'row_id' => $row_id
        ]);

        /* Add new value entry */

        $this->_connection->insert($valueTable, [
            'value_id' => $value_id,
            'row_id' => $row_id,
            'store_id' => 0,
            'label' => $imageFlow->getImageAlt(),
            'position' => $imageFlow->getImagePosition(),
            'disabled' => $imageFlow->getImageHide()
        ]);
    }

    /**
     * Update attributes.
     *
     * @param $row_id
     * @param $value
     * @param Imagesflow $imageFlow
     * @param array $attributesSet
     */
    protected function addAttributes($row_id, $value, Imagesflow $imageFlow, array $attributesSet)
    {
        $assignableAttributes = explode('|', $imageFlow->getImageType());

        if (empty($assignableAttributes)) {
            return;
        }

        $attributesTable = $this->_connection->getTableName('catalog_product_entity_varchar');

        /* Lookup for attibute_code => attribute_id */

        foreach ($assignableAttributes as $attribute) {
            if (!isset($attributesSet[$attribute])) {
                continue;
            }

            $attributeId = ($attributesSet[$attribute])->attribute_id;

            $affectedRows = $this->_connection->insertOnDuplicate($attributesTable, [
                'store_id' => 0,
                'attribute_id' => $attributeId,
                'row_id' => $row_id,
                'value' => $value
            ], ['value']);
        }
    }

    /**
     * @param Imagesflow $image
     * @param string $status
     * @param string $reason
     */
    protected function setProcecessedStatus(ImagesFlow $image, $status = Imagesflow::STATUS_PROCESSED, $reason =null)
    {
        $this->_connection->update(
            $this->_connection->getTableName('flow_imagesflow'),
            [
                'image_process_time' => date('Y-m-d H:i:s'),
                'meta_processed' => 1,
                'image_status'=> $status,
                'log' => $reason
            ],
            $this->_connection->quoteInto('id = ?',$image->getId())
        );
    }

    /**
     * @param $file
     * @return bool
     * @throws \Exception
     */
    public function receiveFile($file)
    {
        if (!$this->_connector) {
            throw new \Exception('no connector joined to channel');
        }

        $meta_file_id = $file->getId();

        $local_dir = $this->_connector->getLocalDir() . "/" . $meta_file_id;

        $this->_connector->setLocalDir($local_dir);

        $model = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\Imagesflow');

        $collection = $model->getCollection();
        $collection->addFieldToFilter('meta_processed', 0);
        $collection->addFieldToFilter('meta_file', $meta_file_id);
        $collection->addFieldToFilter('image_status', 'found');
        $collection->addFieldToFilter('image_queue_time', array('notnull' => true));
        $collection->load();

        $result = true;

        foreach ($collection as $image) {
            try {

                // Is unique filename enabled?
                $unique_filename_pattern = null;
                if ($this->_imagesHelper->isUniqueUploadImagesActive()) {
                    $unique_filename_pattern = "_{$this->unique_image_code_pattern}" . $image->getId();
                }

                $received = $this->_connector->receiveResource($image->getData('image_name'), null, FTP_BINARY, true, $unique_filename_pattern);
                if ($received) {
                    $image->setData('image_receive_time', date("Y-m-d H:i:s"));
                    $image->setData('image_status', 'received');
                    if ($received != $image->getData('image_name')) {
                        $log = $image->getData('log');
                        $log = trim("Image name fixed from '" . $image->getData('image_name') . "' to '{$received}'" . $log . "\n");
                        $image->setData('log', $log);
                        $image->setData('image_name', $received);
                    }
                    $image->save();
                } else {
                    $image->setData('image_status', 'found');
                    $log = $image->getData('log');
                    $log = $log . "\n" . "Image not received";
                    $image->setData('log', $log);
                    $image->save();
                    $result = false;
                }
            } catch (\Exception $e) {
                $log = $image->getData('log');
                $log = $log . "\n" . $e->getMessage();
                $image->setData('log', $log);
                $image->save();
                $result = false;
                //$this->_logger->log('FTP',$e->getMessage());
            }
        }

        //Can't close the connection or the next cycle will fail
        //$this->_connector->close();

        return $result;

    }

    /**
     * @param $file
     * @return bool
     * @throws \Exception
     */
    public function parse($file)
    {
        if (!$this->_connector)
            throw new \Exception('no connector joined to channel');

        $meta_file_id = $file->getId();

        //$local_dir = $this->_connector->getLocalDir().DIRECTORY_SEPARATOR.$meta_file_id;
        //$this->_connector->setLocalDir($local_dir);

        $channel_config = $this->_connector->getConfig();
        $local_dir = $channel_config['local_dir'] . DIRECTORY_SEPARATOR . $meta_file_id;

        //$this->directoryList->getPath('var');

        $csv_filename = $file->getName();
        $channel_name = $file->getChannel();
        $this->loadImageChannelConfig($channel_name);
        $this->filename_separator = $this->getImageSeparator();

        if (!$this->channel_photo_config) {
            throw new \Exception("Wrong images configuration");
        }

        $this->loadImagesListToParse($meta_file_id);

        foreach ($this->collection_of_images_to_parse as $data) {
            $image_filename = $data->getImageName();
            $image_filename_no_ext = $this->removeExtensionFromFilename($image_filename);
            $loaded_data = $this->loadDataBasedOnFilename($image_filename, $channel_name);

            if (count($loaded_data) > 0) {
                $data->setData('image_sku', $loaded_data['sku']);

                //Handling not numeric "L1", "L2", "L3", "L4" special positions (pattern inside filename)
                //Giving custom high order (order is mandatory, numeric and unique)
                switch (strtolower($loaded_data['position'])) {
                    case "l1":
                        $loaded_data['position'] = 999;
                        break;
                    case "l2":
                        $loaded_data['position'] = 998;
                        break;
                    case "l3":
                        $loaded_data['position'] = 997;
                        break;
                    case "l4":
                        $loaded_data['position'] = 996;
                        break;
                    default:
                        #nothing to do
                }

                $data->setData('image_position', (int)$loaded_data['position']);
                if (isset($loaded_data['alt']) && $loaded_data['alt']) {
                    $data->setData('image_alt', $loaded_data['alt']);
                }
                $data->setData('image_type', implode("|", $loaded_data['image_type']));
                $data->setData('image_hide', $loaded_data['image_hide']);
                $data->setData('image_path', $local_dir);
                $data->setData('image_status', 'parsed');
                $data->save();
            } else {
                $data->setData('image_status', 'error');
                $data->save();
                $log = $file->getData('log');
                $log = $log . "\n" . "Can't parse image filename, wrong format! {$csv_filename}";
                $file->setData('log', $log);
                $file->save();
                continue;
            }
        }

        // Handling replicate images for same simple products
        $duplicate_image_super_attribute = trim($this->_imagesHelper->isDuplicateImagesActive());
        if($duplicate_image_super_attribute != ""){
            // Rerunning the same collection
            foreach ($this->collection_of_images_to_parse as $data) {
                if($data->getData('image_status') == self::STATUS_PARSED){
                    $sku_childs_array = $this->_imagesHelper->getAllChildsOfParentWithSameAttribute($data->getData('image_sku'),$duplicate_image_super_attribute);
                    if(count($sku_childs_array) > 0){
                        // For each children replicating the same image to be processed but on a different SKU
                        foreach ($sku_childs_array as $child_sku) {
                            $data->unsetData('id');
                            $data->setData('image_sku',$child_sku);
                            $data->setData('log','Image replicated');
                            $data->save();
                        }
                    }
                }

            }
        }


        return true;
    }

    /**
     * @param $channel_name
     */
    private function loadImageChannelConfig($channel_name)
    {
        $this->channel_photo_config = null;
        if ($this->_parser->loadMap($channel_name)) {
            $config_mapping = $this->_parser->map->getFullMappedData();
            $this->channel_photo_config = $config_mapping;
        }
    }

    /**
     * @return string
     */
    private function getImageSeparator()
    {
        $config = $this->channel_photo_config;

        $has_separator = (isset($config['options']['separator']) && trim($config['options']['separator']) != '');
        return ($has_separator) ? $config['options']['separator'] : "_";

    }

    /**
     * @param $file_id
     */
    private function loadImagesListToParse($file_id)
    {
        $model = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\Imagesflow');

        $collection = $model->getCollection();
        $collection->addFieldToFilter("meta_processed", 0);
        $collection->addFieldToFilter("meta_file", $file_id);
        $collection->addFieldToFilter("image_status", "received");
        $collection->load();

        $this->collection_of_images_to_parse = $collection;

    }

    /**
     * @param $string
     * @return bool|string
     */
    public function removeExtensionFromFilename($string)
    {
        $p = strrpos($string, '.');
        if ($p !== false) {
            $string = substr($string, 0, $p);
        }

        return $string;
    }

    /**
     * @param $image_filename
     * @param null $channel_name
     * @return mixed
     */
    private function loadDataBasedOnFilename($image_filename, $channel_name = null)
    {
        $original_filename = $image_filename;

        // Remove extension
        $image_filename = $this->removeExtensionFromFilename($image_filename);

        $filename_exploded = explode($this->filename_separator, $image_filename);

        $sku_position = $this->getSkuFilenamePosition();
        $sku = "_unknown_";
        if (isset($filename_exploded[$sku_position])) {
            $sku = $filename_exploded[$sku_position];
        }

        $image_position_position = $this->GetImagePositionPosition();

        $position = isset($filename_exploded[$image_position_position]) ? $filename_exploded[$image_position_position] : 1;

        $image_alt_positions = $this->GetImageAltPositions();
        $alt = array();
        if (count($image_alt_positions) > 0) {
            foreach ($image_alt_positions as $pos) {
                // If ALT of image is equal to $this->unique_image_code_pattern then we have an error in the naming and
                // we need to skip that ALT value

                if (isset($filename_exploded[$pos]) && substr($filename_exploded[$pos], 0, 3) != $this->unique_image_code_pattern) {
                    $alt[] = $filename_exploded[$pos];
                }

            }
            if (count($alt) > 0) {
                $alt = implode($this->image_alt_separator, $alt);
            }
        }

        if ($alt == array()) {
            $alt = "";
        }

        $image_type = $this->getImageType($position);
        $image_hide = $this->getImageHide($position);

        $this->array_of_data_for_csv[$sku][$position]['filename'] = $original_filename;
        $this->array_of_data_for_csv[$sku][$position]['path'] = $this->_connector->getLocalDir();
        $this->array_of_data_for_csv[$sku][$position]['position'] = $position;
        if ($alt) {
            $this->array_of_data_for_csv[$sku][$position]['alt'] = $alt;
        }
        $this->array_of_data_for_csv[$sku][$position]['image_type'] = $image_type;
        $this->array_of_data_for_csv[$sku][$position]['image_hide'] = $image_hide;
        $this->array_of_data_for_csv[$sku][$position]['sku'] = $sku;

        // Start dispatch event to handle change of this data if needed
        $event_dispatch_data = new \Magento\Framework\DataObject();
        $event_dispatch_data->setData('channel_name', $channel_name);
        $event_dispatch_data->setData('image_filename', $original_filename);
        $event_dispatch_data->setData('image_filename_noext', $image_filename);
        $event_dispatch_data->setData('filename_separator', $this->filename_separator);
        $event_dispatch_data->setData('channel_mapping', $this->channel_photo_config);
        $event_dispatch_data->setData('loaded_data', $this->array_of_data_for_csv[$sku][$position]);
        $this->eventManager->dispatch('flow_images_parse_load_data_based_on_filename_after', ['event_dispatch_data' => $event_dispatch_data]);
        // If the value is changed then it will be updated here
        $this->array_of_data_for_csv[$sku][$position] = $event_dispatch_data->getData('loaded_data');
        // End dispatch event, the data may have been changed here

        return $this->array_of_data_for_csv[$sku][$position];
    }

    /**
     * @return int|string
     */
    private function getSkuFilenamePosition()
    {
        $config = $this->channel_photo_config;
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] == 'sku') {
                return $key;
            }
        }

        //Handle error missing SKU on config here, block all flow
    }

    /**
     * @return int|string
     */
    private function GetImagePositionPosition()
    {
        $config = $this->channel_photo_config;
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] == 'position') {
                return $key;
            }
        }

        //Handle error missing SKU on config here, block all flow
    }

    /**
     * @return array
     */
    private function GetImageAltPositions()
    {
        $config = $this->channel_photo_config;
        $positions = array();
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] == 'alt') {
                $positions[] = $key;
            }
        }

        return $positions;

        //Handle error missing SKU on config here, block all flow
    }

    /**
     * @param $position
     * @return array
     * This will look into the json config of the image channels
     * searching which type of image is, based on the $position key inside the "types" node of the json
     */
    private function getImageType($position)
    {
        $position = trim($position);

        $config = $this->channel_photo_config;

        if (isset($config['types'][$position]['type'])) {
            return $config['types'][$position]['type'];
        }

        // Check if position is a number (can also be a string)
        // If is a number handle it as a number
        // This will handle positions as "001" "01" "1" (usually human errors)
        if((int)$position > 0 && isset($config['types'][(int)$position]['type'])){
            return $config['types'][(int)$position]['type'];
        }

        // Check again if maybe we have uppercases (usually human errors)
        $position = strtolower($position);
        if (isset($config['types'][$position]['type'])) {
            return $config['types'][$position]['type'];
        }

        return array();
    }

    /**
     * @param $position
     * @return int
     */
    private function getImageHide($position)
    {
        $config = $this->channel_photo_config;

        if (isset($config['types'][$position]['hide']) && $config['types'][$position]['hide'] == "true") {
            return 1;
        } else {
            return 0;
        }

    }

    /*
     * $filename is column 'value' from 'catalog_product_entity_media_gallery'
     */

    /**
     * @param $path
     * @return bool
     */
    private function removeCatalogProductImageByPath($path)
    {
        $file = BP . "/pub/media/catalog/product{$path}";
        $pathinfo = pathinfo($file);
        $filename = $pathinfo['basename'];

        $resource = \Magento\Framework\App\ObjectManager::getInstance()->create('\Magento\Framework\App\ResourceConnection');
        $connection = $resource->getConnection(\Magento\Framework\App\ResourceConnection::DEFAULT_CONNECTION);

        $query = "DELETE FROM catalog_product_entity_media_gallery WHERE value = '{$path}'";
        $result = $connection->query($query);

        if (!$result) {
            $this->_logger->info("INSERT IMAGES: NOT Removed from catalog_product_entity_media_gallery {$path}");
            return false;
        } else {
            $this->_logger->info("INSERT IMAGES: Removed from catalog_product_entity_media_gallery {$path}");
        }

        $ProductMediaImageAttributeIds = $this->getProductMediaImageAttributeIds();
        if ($ProductMediaImageAttributeIds) {
            $add_where = " AND attribute_id IN ({$ProductMediaImageAttributeIds}) ";
        }

        // Removing also the "photo tags" (base, small, thumbnail) or they will be left there
        $query = "DELETE FROM catalog_product_entity_varchar WHERE value = '{$path}' {$add_where}";
        $result = $connection->query($query);

        if (!$result) {
            $this->_logger->info("INSERT IMAGES: NOT Removed from catalog_product_entity_varchar {$path}");
            return false;
        } else {
            $this->_logger->info("INSERT IMAGES: Removed TAG from catalog_product_entity_varchar {$path}");
        }

        if (is_file($file)) {
            unlink($file);
            $this->_logger->info("INSERT IMAGES: Removed file from disk {$file}");
        } else {
            $this->_logger->info("INSERT IMAGES: Removing file from disk {$file} not found, skipping");
        }

        $this->removeImageCacheFromImageFileName($path);

        return true;
    }

    /**
     * @return string of ids comma separated
     */
    private function getProductMediaImageAttributeIds()
    {
        /** @var  $coll \Magento\Eav\Model\ResourceModel\Entity\Attribute\Collection */
        $coll = \Magento\Framework\App\ObjectManager::getInstance()->create(\Magento\Eav\Model\ResourceModel\Entity\Attribute\Collection::class);
        $coll->addFieldToFilter('frontend_input', 'media_image');
        $attrAll = $coll->load()->getItems();
        $result = array();
        foreach ($attrAll as $attr) {
            $result[] = $attr->getAttributeId();
        }

        $return = false;
        if (count($result)) {
            $return = implode(",", $result);
        }

        return $return;
    }

    /**
     * @param null $filename
     * @return bool
     */
    private function removeImageCacheFromImageFileName($filename = null)
    {

        if (!$this->_imagesHelper->isClearImageCacheActive() || !$filename) {
            return true;
        }

        // Array of all cache main folders
        try {
            $main_cache_image_paths = $this->getAllCacheImageMainPaths();
        } catch (\Exception $e) {
            $this->_logger->info("ERROR IMAGES: removeImageCacheFromImageFileName " . $e->getMessage());
            return false;
        }

        if (!isset($main_cache_image_paths) || !$main_cache_image_paths || count($main_cache_image_paths) == 0) {
            return false;
        }

        foreach ($main_cache_image_paths as $path) {

            // Safe check
            $path = rtrim($path, DIRECTORY_SEPARATOR);

            $full_path = $this->cache_folder . $path . $filename;

            if (file_exists($full_path)) {
                // Try to delete the cache
                unlink($full_path);
                $this->_logger->info("INSERT IMAGES: Removed cache image {$full_path}");
            }
        }

    }

    /**
     * @return array
     */
    public function getAllCacheImageMainPaths()
    {
        // Give me an array of all the folders inside the cache folder
        // Going deep 3
        $image_cache_paths_array = $this->_imagesHelper->foldersMap($this->cache_folder, 3);

        // Transform the array paths into string path checking also the md5 of the last folder
        $array_of_builded_paths = array();
        $this->_imagesHelper->collapse("", $image_cache_paths_array, $array_of_builded_paths);

        return $array_of_builded_paths;
    }
}
